from typing import Tuple, Optional, List, Dict, cast, Union

import numpy as np

from ..constants import CellValue, tileProduction, ItemProperty
from ..constants.Direction import transformCoords, Direction
from ..constants.Resource import Resource
from ..state import City, Player
from ..state.GameState import GameState


class Rules:

    def __init__(self, state: GameState):
        self.__state = state

    @property
    def state(self) -> GameState:
        return self.__state

    def getCityTopLeft(self, cell: Tuple[int, int]) -> Tuple[int, int]:
        layer = self.__state.world.buildings
        value = layer.getValue(cell)
        if value != CellValue.BUILDINGS_CITY:
            raise ValueError(f"No city at {cell}")
        item = layer.getItem(cell)
        tlCell = transformCoords(cell, Direction.TOPLEFT)
        if layer.contains(tlCell) and layer.getValue(tlCell) == value:
            assert layer.getItem(tlCell) == item
            return tlCell
        tlCell = transformCoords(cell, Direction.TOP)
        if layer.contains(tlCell) and layer.getValue(tlCell) == value:
            assert layer.getItem(tlCell) == item
            return tlCell
        tlCell = transformCoords(cell, Direction.LEFT)
        if layer.contains(tlCell) and layer.getValue(tlCell) == value:
            assert layer.getItem(tlCell) == item
            return tlCell
        return cell

    def findCity(self, cityOrCell: Union[City, Tuple[int, int]]) -> Optional[City]:
        world = self.__state.world
        if isinstance(cityOrCell, City):
            city = cast(City, cityOrCell)
        elif type(cityOrCell) == tuple:
            buildings = world.buildings
            item = buildings.getItem(cityOrCell)
            if not isinstance(item, City):
                return None
            city = cast(City, item)
        else:
            return None
        return city

    def getCity(self, cityOrCell: Union[City, Tuple[int, int]]) -> City:
        city = self.findCity(cityOrCell)
        if city is None:
            raise ValueError(f"Invalid city")
        return city

    def getCityHouseArea(self, cityCell: Tuple[int, int], valueIn: int = 1, valueOut: int = 0) -> \
            Tuple[Tuple[int, int], np.ndarray]:
        cityCell = self.getCityTopLeft(cityCell)
        area = Rules.houseArea.copy() * valueIn
        if valueOut != 0:
            area += (1 - Rules.houseArea) * valueOut
        pos = (
            cityCell[0] - area.shape[0] // 2 + 1,
            cityCell[1] - area.shape[1] // 2 + 1
        )
        return pos, area

    def getCityArea(self, cityCell: Tuple[int, int], valueIn: int = 1, valueOut: int = 0) \
            -> Tuple[Tuple[int, int], np.ndarray]:
        cityCell = self.getCityTopLeft(cityCell)
        if Rules.cityArea is None:
            import math
            radius = 8

            def circle(i, j):
                x = i - radius + 0.5
                y = j - radius + 0.5
                return 1 if math.sqrt(x * x + y * y) < radius else 0

            Rules.cityArea = np.array(
                [[circle(i, j) for i in range(2 * radius)] for j in range(2 * radius)],
                dtype=np.int32
            )

        area = Rules.cityArea.copy() * valueIn
        if valueOut != 0:
            area += (1 - Rules.cityArea) * valueOut
        pos = (
            cityCell[0] - area.shape[0] // 2 + 1,
            cityCell[1] - area.shape[1] // 2 + 1
        )
        return pos, area

    def isInCityArea(self, cityCell: Tuple[int, int], cell: Tuple[int, int]) -> bool:
        pos, cityArea = self.getCityArea(cityCell)
        posX, posY = pos
        x, y = cell
        if x < posX or x >= posX + cityArea.shape[0]:
            return False
        if y < posY or y >= posY + cityArea.shape[1]:
            return False
        x -= posX
        y -= posY
        return cityArea[x, y] != 0

    def computeCityzenCount(self, cityCell: Tuple[int, int]):
        pos, houseArea = self.getCityHouseArea(cityCell)
        objects = self.__state.world.objects
        _, mask = objects.compare(pos, CellValue.OBJECTS_HOUSES, houseArea)
        return mask.sum()

    def findCityFreeHouses(self, cityCell: Tuple[int, int]) -> List[Tuple[int, int]]:
        pos, houseArea = self.getCityHouseArea(cityCell)

        world = self.__state.world
        ground = world.ground
        impassable = world.impassable
        objects = world.objects
        buildings = world.buildings

        intersection = ground.computeAreaIntersection(pos, houseArea)
        if intersection is None:
            return []
        cells = []
        sourceMinX, sourceMinY, destMinX, destMinY, width, height = intersection
        for y in range(height):
            for x in range(width):
                sourceX, sourceY = sourceMinX + x, sourceMinY + y
                if houseArea[sourceX, sourceY] == 0:
                    continue
                dest = destMinX + x, destMinY + y
                if ground.getValue(dest) != CellValue.GROUND_EARTH:
                    continue
                if impassable.getValue(dest) != CellValue.NONE:
                    continue
                if objects.getValue(dest) != CellValue.NONE:
                    continue
                if buildings.getValue(dest) != CellValue.NONE:
                    continue
                cells.append(dest)

        return cells

    def computeCityWorkersProduction(self, cityCell: Tuple[int, int], sortProduction: Optional[Resource]) -> \
            List[Tuple[Tuple[int, int], Dict[Resource, int]]]:
        cityCell = self.getCityTopLeft(cityCell)
        city = self.getCity(cityCell)
        cells: List[Tuple[Tuple[int, int], Dict[Resource, int]]] = []
        for cell in city.getWorkersCell():
            resources = self.computeCellResources(cityCell, cell)
            production: Dict[Resource, int] = {}
            addResources(production, resources)
            cells.append((cell, production))

        if sortProduction:
            def sortFunc(x: Tuple[Tuple[int, int], Dict[Resource, int]]):
                cell, prod = x[0], x[1]
                if sortProduction in prod:
                    dist = (cell[0] - cityCell[0] - 0.5) ** 2 + (cell[1] - cityCell[1] - 0.5) ** 2
                    return (prod[sortProduction] * 10000 + 10000 - dist) * 10 + (3 * cell[0] + 7 * cell[1]) % 10
                return 0

            cells.sort(key=sortFunc)

        return cells

    def findCityUnassignedCells(self, cityCell: Tuple[int, int], sortProduction: Optional[Resource]) -> \
            List[Tuple[Tuple[int, int], Dict[Resource, int]]]:
        cityCell = self.getCityTopLeft(cityCell)
        pos, area = self.getCityArea(cityCell)
        housesPos, housesArea = self.getCityHouseArea(cityCell)
        housesWidth, housesHeight = housesArea.shape[0], housesArea.shape[1]
        city = self.getCity(cityCell)
        workersCell = city.getWorkersCell()
        cells: List[Tuple[Tuple[int, int], Dict[Resource, int]]] = []
        for y in range(area.shape[1]):
            for x in range(area.shape[0]):
                if area[x, y] == 0:
                    continue
                cell = pos[0] + x, pos[1] + y
                hX = cell[0] - housesPos[0]
                hY = cell[1] - housesPos[1]
                if 0 <= hX < housesWidth and 0 <= hY < housesHeight:
                    if housesArea[hX, hY] != 0:
                        continue
                if cell in workersCell:
                    continue
                resources = self.computeCellResources(cityCell, cell)
                if not resources:
                    continue
                production: Dict[Resource, int] = {}
                addResources(production, resources)
                cells.append((cell, production))

        if sortProduction:
            def sortFunc(x: Tuple[Tuple[int, int], Dict[Resource, int]]):
                cell, prod = x[0], x[1]
                if sortProduction in prod:
                    dist = (cell[0] - cityCell[0] - 0.5) ** 2 + (cell[1] - cityCell[1] - 0.5) ** 2
                    return (prod[sortProduction] * 10000 + 10000 - dist) * 10 + (3 * cell[0] + 7 * cell[1]) % 10
                return 0

            cells.sort(key=sortFunc, reverse=True)

        return cells

    def computeCellResources(self, cityOrCell: Union[City, Tuple[int, int]], cell: Tuple[int, int]) -> List[Resource]:
        world = self.__state.world
        if not world.contains(cell):
            return []
        buildingValue = world.buildings.getValue(cell)
        if buildingValue != CellValue.NONE and buildingValue != CellValue.BUILDINGS_ASSIGNED:
            return []
        groundValue = CellValue(world.ground.getValue(cell))
        impassableValue = CellValue(world.impassable.getValue(cell))
        objectsValue = CellValue(world.objects.getValue(cell))
        if groundValue in tileProduction:
            impassableProduction = tileProduction[groundValue]
            if impassableValue in impassableProduction:
                objectsProduction = impassableProduction[impassableValue]
                if objectsValue in objectsProduction:
                    return objectsProduction[objectsValue]
        return []

    def getCityGranarySize(self, cityOrCell: Union[City, Tuple[int, int]]):
        city = self.getCity(cityOrCell)
        return city.getProperty(ItemProperty.GRANARY_FOOD_MAX)

    def getCityGrowthMax(self, cityOrCell: Union[City, Tuple[int, int]]):
        city = self.getCity(cityOrCell)
        citizenCount = city.getCitizenCount()
        growthMin = cast(int, city.getProperty(ItemProperty.GROWTH_POINTS_MIN, 1))
        growthRate = cast(int, city.getProperty(ItemProperty.GROWTH_POINTS_RATE, 1))
        growthMax = cast(int, city.getProperty(ItemProperty.GROWTH_POINTS_MAX, 10))
        return min(max(int(citizenCount / growthRate), growthMin), growthMax)

    def computeCityProduction(self, cityOrCell: Union[City, Tuple[int, int]]) -> \
            Dict[str, Dict[Resource, int]]:
        city = self.getCity(cityOrCell)
        base: Dict[Resource, int] = {}
        addResources(base, city.getResourcesProperty(ItemProperty.BASE_PRODUCTION, []))
        total = base.copy()

        workers: Dict[Resource, int] = {}
        for worker in city.getWorkers():
            assert worker.cell is not None
            resources = self.computeCellResources(city, worker.cell)
            addResources(workers, resources)
            addResources(total, resources)

        merchants: Dict[Resource, int] = {}
        merchantProduction = city.getResourcesProperty(ItemProperty.MERCHANT_PRODUCTION, [])
        for merchant in city.getMerchants():
            addResources(workers, merchantProduction)
            addResources(total, merchantProduction)

        balance = total.copy()
        if Resource.FOOD not in balance:
            balance[Resource.FOOD] = 0
        balance[Resource.FOOD] -= 2 * city.getCitizenCount()

        return {
            "base": base,
            "merchants": merchants,
            "workers": workers,
            "balance": balance,
            "total": total,
        }

    def getCities(self) -> List[Tuple[Tuple[int, int], City]]:
        cities = []
        buildings = self.__state.world.buildings
        for itemCells in buildings.getItems():
            if not isinstance(itemCells.item, City):
                continue
            city = cast(City, itemCells.item)
            cell = self.getCityTopLeft(itemCells.cells[0])
            cities.append((cell, city))
        return cities

    def getPlayerCities(self, playerOrId: Union[Player, int]) -> List[Tuple[Tuple[int, int], City]]:
        cities = []
        if isinstance(playerOrId, Player):
            playerId = playerOrId.id
        else:
            playerId = int(playerOrId)
        buildings = self.__state.world.buildings
        for itemCells in buildings.getPlayerItems(playerId):
            if not isinstance(itemCells.item, City):
                continue
            city = cast(City, itemCells.item)
            cell = self.getCityTopLeft(itemCells.cells[0])
            cities.append((cell, city))
        return cities

    def computePlayerProduction(self, playerOrId: Union[Player, int]) -> Dict[str, Dict[Resource, int]]:
        balance: Dict[Resource, int] = {}
        for city in self.getPlayerCities(playerOrId):
            cityProduction = self.computeCityProduction(city[1])
            addResources(balance, cityProduction["balance"])
        return {
            "balance": balance
        }

    houseArea = np.array([
        [0, 1, 1, 1, 1, 0],
        [1, 1, 1, 1, 1, 1],
        [1, 1, 1, 1, 1, 1],
        [1, 1, 1, 1, 1, 1],
        [1, 1, 1, 1, 1, 1],
        [0, 1, 1, 1, 1, 0],
    ], dtype=np.int32)

    cityArea: Optional[np.ndarray] = None


def addResources(total: Dict[Resource, int], resources: Union[List[Resource], Dict[Resource, int]]):
    if type(resources) == list:
        for resource in resources:
            if resource in total:
                total[resource] += 1
            else:
                total[resource] = 1
    elif type(resources) == dict:
        for resource, count in resources.items():
            if resource in total:
                total[resource] += count
            else:
                total[resource] = count
    else:
        raise ValueError(f"Invalid type {type(resources)}")
