from __future__ import annotations

import logging

from core.constants import ItemProperty, Resource, CellValue
from core.logic.Command import Command, STATE_PRIORITY
from core.logic.Logic import Logic
from core.logic.command import AssignCitizen


class NextTurn(Command):

    def priority(self) -> int:
        return STATE_PRIORITY

    def check(self, logic: Logic) -> bool:
        state = logic.state
        if state.playerId not in state.playerIds:
            logging.error("Current player id not in state player list!")
            return False
        return True

    def execute(self, logic: Logic):
        state = logic.state

        # Next player id / turn
        playerIds = state.playerIds
        playerIndex = playerIds.index(state.playerId) + 1
        if playerIndex >= len(playerIds):
            playerIndex = 0
            self.updateAll(logic)

        state.playerId = playerIds[playerIndex]

        # Next turn
        state.notifyTurnChanged()

    def updateAll(self, logic: Logic):
        state = logic.state
        rules = logic.rules

        # Reset unit points
        items = state.world.units.items
        for unit in items.values():
            if unit.hasProperty(ItemProperty.MOVE_POINTS):
                points = unit.getIntProperty(ItemProperty.MAX_MOVE_POINTS)
                unit.setIntProperty(ItemProperty.MOVE_POINTS, points)
            if unit.hasProperty(ItemProperty.ACTION_POINTS):
                points = unit.getIntProperty(ItemProperty.MAX_ACTION_POINTS)
                unit.setIntProperty(ItemProperty.ACTION_POINTS, points)

        # Update resources
        for player in state.players:
            production = rules.computePlayerProduction(player)
            for resource, amount in production["balance"].items():
                if resource == Resource.FOOD:
                    continue
                player.addResource(resource, amount)

        # Update cities
        for cityCell, city in rules.getCities():
            self.updateCity(logic, cityCell, city)

        state.turn = state.turn + 1

    def updateCity(self, logic: Logic, cityCell, city):
        rules = logic.rules
        world = logic.world

        production = rules.computeCityProduction(city)
        foodBalance = production["balance"][Resource.FOOD]
        granary = city.getProperty(ItemProperty.GRANARY_FOOD, 0)
        granaryMax = rules.getCityGranarySize(city)
        granary = min(granaryMax, granary + foodBalance)
        if granary >= 0:
            city.setProperty(ItemProperty.GRANARY_FOOD, granary)

            growth = city.getProperty(ItemProperty.GROWTH_POINTS, 0) + 1
            growthMax = rules.getCityGrowthMax(city)
            if growth < growthMax:
                city.setProperty(ItemProperty.GROWTH_POINTS, growth)
                return
            freeCells = rules.findCityFreeHouses(cityCell)
            if not freeCells:
                return

            randomSeed = (3 * cityCell[0] + 7 * cityCell[1]) % len(freeCells)
            freeCell = freeCells[randomSeed]
            world.objects.setValue(freeCell, CellValue.OBJECTS_HOUSES)
            world.objects.notifyCellChanged(freeCell)

            city.addCitizen()
            unassignedCells = rules.findCityUnassignedCells(cityCell, Resource.FOOD)
            if unassignedCells:
                index = 0
                assert city.findWorker(unassignedCells[index][0]) is None
                unassignedCell = unassignedCells[index][0]
                command = AssignCitizen(cityCell, unassignedCell)
                if command.check(logic):
                    command.execute(logic)

            city.setProperty(ItemProperty.GROWTH_POINTS, 0)
        else:
            city.setProperty(ItemProperty.GROWTH_POINTS, 0)
            city.setProperty(ItemProperty.GRANARY_FOOD, 0)

            workersCell = rules.computeCityWorkersProduction(cityCell, Resource.FOOD)
            if not workersCell:
                return
            city.removeWorker(workersCell[0][0])


