from __future__ import annotations

from typing import Tuple, Optional, List, cast

from core.constants import CellValue, ItemProperty
from core.logic import DistanceMap
from core.logic.Command import Command, WORLD_PRIORITY, WORLD_MAX_WIDTH
from core.logic.Logic import Logic
from core.state import Unit


class MoveUnit(Command):

    def __init__(self, fromCell: Tuple[int, int], toCell: Tuple[int, int],
                 distanceMap: Optional[DistanceMap] = None):
        self.__fromCell = fromCell
        self.__toCell = toCell
        self.__path: List[Tuple[int, int]] = []
        self.__costs: List[int] = []
        self.__distanceMap = distanceMap

    @property
    def distanceMap(self) -> Optional[DistanceMap]:
        return self.__distanceMap

    @property
    def path(self) -> List[Tuple[int, int]]:
        return self.__path

    @property
    def costs(self) -> List[int]:
        return self.__costs

    def priority(self) -> int:
        return WORLD_PRIORITY + self.__fromCell[0] + self.__fromCell[1] * WORLD_MAX_WIDTH

    def check(self, logic: Logic) -> bool:
        if self.__fromCell == self.__toCell:
            return False

        world = logic.world
        if not world.contains(self.__fromCell):
            return False
        if not world.contains(self.__toCell):
            return False

        item = world.units.getItem(self.__fromCell)
        if item is None:
            return False
        unit = cast(Unit, item)
        if unit.playerId != logic.state.playerId:
            return False

        # Compute best direction (if any)
        if self.__distanceMap is None:
            fromX, fromY = self.__fromCell
            toX, toY = self.__toCell
            radius = 10
            area = (
                min(fromX, toX) - radius, min(fromY, toY) - radius,
                max(fromX, toX) + radius + 1, max(fromY, toY) + radius + 1
            )
            self.__distanceMap = DistanceMap(world, area)
            self.__distanceMap.compute(self.__fromCell)
        self.__path, self.__costs = self.__distanceMap.getPath(self.__toCell)
        if len(self.__path) == 0:
            return False

        movePoints = unit.getIntProperty(ItemProperty.MOVE_POINTS)
        if self.__costs[0] > movePoints:
            return False

        return True

    def execute(self, logic: Logic):
        unitsLayer = logic.world.units
        unit = unitsLayer.getItem(self.__fromCell)
        if unit is None:
            raise RuntimeError("No unit")
        unitsLayer.setItem(self.__fromCell, CellValue.NONE)
        targetCell = self.__path[-1]
        unitsLayer.setItem(targetCell, CellValue.UNITS_UNIT, unit)

        cost = self.__costs[-1]
        movePoints = unit.getIntProperty(ItemProperty.MOVE_POINTS)
        movePoints -= cost
        unit.setIntProperty(ItemProperty.MOVE_POINTS, movePoints)

        unitsLayer.notifyCellChanged(self.__fromCell)
        unitsLayer.notifyCellChanged(targetCell)
        logic.state.notifyUnitMoved(self.__fromCell, self.__toCell)

        if len(self.__path) > 1:
            command = MoveUnit(targetCell, self.__toCell)
            logic.addCommand(command)

