from typing import List, Tuple, Dict, Any, cast

from core.state.Layer import Layer
from ..IDataTransfer import IDataTransfer
from ..constants import CellValue


class World(IDataTransfer):
    def __init__(self, size: Tuple[int, int]):
        self.__size = size
        self.__layers = {
            "ground": Layer(size, CellValue.GROUND_SEA),
            "impassable": Layer(size, CellValue.NONE),
            "objects": Layer(size, CellValue.NONE),
            "buildings": Layer(size, CellValue.NONE),
            "units": Layer(size, CellValue.NONE),
        }

    def __eq__(self, other) -> bool:
        if not isinstance(other, World):
            return False
        world = cast(World, other)
        if self.__size != world.__size:
            return False
        if len(self.__layers) != len(world.__layers):
            return False
        for name, layer in self.__layers.items():
            if name not in world.__layers:
                return False
            if layer != world.__layers[name]:
                return False
        return True

    @property
    def size(self) -> Tuple[int, int]:
        return self.__size

    @property
    def width(self) -> int:
        return self.__size[0]

    @property
    def height(self) -> int:
        return self.__size[1]

    def contains(self, coords: Tuple[int, int]) -> bool:
        return 0 <= coords[0] < self.__size[0] and 0 <= coords[1] < self.__size[1]

    @property
    def ground(self) -> Layer:
        return self.__layers["ground"]

    @property
    def impassable(self) -> Layer:
        return self.__layers["impassable"]

    @property
    def objects(self) -> Layer:
        return self.__layers["objects"]

    @property
    def buildings(self) -> Layer:
        return self.__layers["buildings"]

    @property
    def units(self) -> Layer:
        return self.__layers["units"]

    @property
    def layerNames(self) -> List[str]:
        return list(self.__layers.keys())

    @property
    def layers(self) -> List[Layer]:
        return list(self.__layers.values())

    def __getitem__(self, name: str) -> Layer:
        return self.getLayer(name)

    def getLayer(self, name: str) -> Layer:
        if name not in self.__layers:
            raise ValueError(f"No layer {name}")
        return self.__layers[name]

    def gatherData(self) -> Dict[str, Any]:
        layersData = {}
        for name, layer in self.__layers.items():
            layersData[name] = layer.gatherData()
        return {
            "width": int(self.__size[0]),
            "height": int(self.__size[1]),
            "layers": layersData
        }

    def takeData(self, data: Dict[str, Any]):
        self.__size = (int(data["width"]), int(data["height"]))
        self.__layers.clear()
        for name, layerData in data["layers"].items():
            layer = Layer(self.__size)
            layer.takeData(layerData)
            self.__layers[name] = layer
