import copy
from typing import Dict, Any, cast, List, Tuple, Optional

from core.constants import ItemProperty, CityProperties, CityClass
from .Citizen import Citizen
from .Item import Item


class City(Item):
    __slots__ = [
        "__cityClass", "__citizens", "__name"
    ]

    def __init__(self, cityClass: CityClass = CityClass.DEFAULT, playerId: int = 0):
        super().__init__(playerId)
        self.__citizens: List[Citizen] = []
        self.__name = "NoName"
        self.__cityClass = cityClass
        if cityClass not in CityClass:
            raise ValueError(f"Invalid class {cityClass}")
        self._properties = copy.deepcopy(CityProperties[cityClass])

    def __eq__(self, other) -> bool:
        if not isinstance(other, City):
            return False
        city = cast(City, other)
        if self.__name != city.__name:
            return False
        if self.__citizens != city.__citizens:
            return False
        if self.__cityClass != city.__cityClass:
            return False
        return super().__eq__(other)

    def __repr__(self) -> str:
        return f"City {self.__name}"

    def getDefaultProperties(self) -> Dict[ItemProperty, Any]:
        return CityProperties[self.__cityClass]

    @property
    def name(self) -> str:
        return self.__name

    def clearCitizens(self):
        self.__citizens.clear()

    def addCitizen(self, citizen: Optional[Citizen] = None):
        if citizen is None:
            citizen = Citizen()
        self.__citizens.append(citizen)

    def getCitizenCount(self) -> int:
        return len(self.__citizens)

    def getWorkerCount(self) -> int:
        return sum(1 for citizen in self.__citizens if citizen.cell is not None)

    def getWorkers(self) -> List[Citizen]:
        return [citizen for citizen in self.__citizens if citizen.cell is not None]

    def getMerchantCount(self) -> int:
        return sum(1 for citizen in self.__citizens if citizen.cell is None)

    def getMerchants(self) -> List[Citizen]:
        return [citizen for citizen in self.__citizens if citizen.cell is None]

    def getCitizen(self, index: int) -> Citizen:
        return self.__citizens[index]

    def getWorkersCell(self) -> Dict[Tuple[int, int], Citizen]:
        return {
            citizen.cell: citizen for citizen in self.__citizens if citizen.cell is not None
        }

    def findWorker(self, cell: Tuple[int, int]) -> Optional[Citizen]:
        for citizen in self.__citizens:
            if citizen.cell == cell:
                return citizen
        return None

    def removeWorker(self, cell: Tuple[int, int]):
        found = -1
        for index, citizen in enumerate(self.__citizens):
            if citizen.cell == cell:
                found = index
                break
        if found >= 0:
            del self.__citizens[found]

    def assignCitizen(self, cell: Tuple[int, int]):
        for citizen in self.__citizens:
            if citizen.cell is None:
                citizen.cell = cell
                return
        raise RuntimeError("No more unassigned citizen")

    def unassignCitizen(self, cell: Tuple[int, int]):
        for citizen in self.__citizens:
            if citizen.cell == cell:
                citizen.cell = None
                return

    # Data transfer

    def gatherData(self) -> Dict[str, Any]:
        data = super().gatherData()
        data["type"] = "city"
        data["cityClass"] = self.__cityClass.toName()
        data["name"] = self.__name
        data["citizens"] = [citizen.gatherData() for citizen in self.__citizens]
        return data

    def takeData(self, data: Dict[str, Any]):
        super().takeData(data)
        self.__cityClass = CityClass.fromName(data['cityClass'])
        self.__name = data["name"]
        self.__citizens = [Citizen(citizenData) for citizenData in data["citizens"]]
