from typing import Tuple

from core.constants import Resource, ItemProperty
from core.logic import Logic
from core.state import GameState, IGameStateListener, City
from .FrameComponent import FrameComponent
from ..control import Label
from ...theme.Theme import Theme
from ...tools import formatResourceBalance


class CityFrame(FrameComponent, IGameStateListener):

    def __init__(self, theme: Theme, logic: Logic, city: City):
        super().__init__(theme)
        self.__logic = logic
        self.__city = city

        self.__nameLabel = Label(theme, city.name)
        self.__nameLabel.moveRelativeTo("topLeft", self, "topLeft")
        self.addComponent(self.__nameLabel)

        self.__citizensLabel = Label(theme)
        self.__workersLabel = Label(theme)
        self.__merchantsLabel = Label(theme)
        self.addComponent(self.__citizensLabel)
        self.addComponent(self.__workersLabel)
        self.addComponent(self.__merchantsLabel)

        self.__foodLabel = Label(theme)
        self.__granaryLabel = Label(theme)
        self.__growthLabel = Label(theme)
        self.addComponent(self.__foodLabel)
        self.addComponent(self.__granaryLabel)
        self.addComponent(self.__growthLabel)

        self.__woodLabel = Label(theme)
        self.__stoneLabel = Label(theme)
        self.__goldLabel = Label(theme)
        self.addComponent(self.__woodLabel)
        self.addComponent(self.__stoneLabel)
        self.addComponent(self.__goldLabel)

        self.__updateValues()

        # Listeners
        self.__logic.state.registerListener(self)

    def dispose(self):
        self.__logic.state.removeListener(self)

    # GameState listener

    def citizenAssigned(self, state: GameState, cityCell: Tuple[int, int], cell: Tuple[int, int]):
        self.__updateValues()

    def citizenUnassigned(self, state: GameState, cityCell: Tuple[int, int], cell: Tuple[int, int]):
        self.__updateValues()

    def turnChanged(self, state: GameState):
        self.__updateValues()

    def resourcesChanged(self, state: GameState):
        self.__updateValues()

    def __updateValues(self):
        city = self.__city
        rules = self.__logic.rules

        self.__citizensLabel.setMessage(f"{city.getCitizenCount()}<citizen>")
        self.__workersLabel.setMessage(f"{city.getWorkerCount()}<worker>")
        self.__merchantsLabel.setMessage(f"{city.getMerchantCount()}<merchant>")

        production = rules.computeCityProduction(city)
        balance = production["balance"]

        self.__foodLabel.setMessage(f"{formatResourceBalance(Resource.FOOD, balance)}<food>")
        granaryFood = city.getProperty(ItemProperty.GRANARY_FOOD, 0)
        granaryFoodMax = rules.getCityGranarySize(city)
        self.__granaryLabel.setMessage(f"{granaryFood}/{granaryFoodMax}<granary>")
        growthFood = city.getProperty(ItemProperty.GROWTH_POINTS, 0)
        growthFoodMax = rules.getCityGrowthMax(city)
        self.__growthLabel.setMessage(f"{growthFood}/{growthFoodMax}<growth>")

        self.__woodLabel.setMessage(f"{formatResourceBalance(Resource.WOOD, balance)}<wood>")
        self.__stoneLabel.setMessage(f"{formatResourceBalance(Resource.STONE, balance)}<stone>")
        self.__goldLabel.setMessage(f"{formatResourceBalance(Resource.GOLD, balance)}<gold>")

        borderSize = self.theme.framePadding
        self.__citizensLabel.moveRelativeTo("topLeft", self.__nameLabel, "bottomLeft", borderSize=borderSize)
        self.__workersLabel.moveRelativeTo("topLeft", self.__citizensLabel, "topRight", borderSize=borderSize)
        self.__merchantsLabel.moveRelativeTo("topLeft", self.__workersLabel, "topRight", borderSize=borderSize)

        self.__foodLabel.moveRelativeTo("topLeft", self.__citizensLabel, "bottomLeft", borderSize=borderSize)
        self.__granaryLabel.moveRelativeTo("topLeft", self.__foodLabel, "topRight", borderSize=borderSize)
        self.__growthLabel.moveRelativeTo("topLeft", self.__granaryLabel, "topRight", borderSize=borderSize)

        self.__woodLabel.moveRelativeTo("topLeft", self.__foodLabel, "bottomLeft", borderSize=borderSize)
        self.__stoneLabel.moveRelativeTo("topLeft", self.__woodLabel, "topRight", borderSize=borderSize)
        self.__goldLabel.moveRelativeTo("topLeft", self.__stoneLabel, "topRight", borderSize=borderSize)

        self.pack()
