from typing import Tuple, cast, Optional

from core.constants import ItemProperty
from core.logic import Logic, Command
from core.logic.command import AssignCitizen, UnassignCitizen
from core.state import IGameStateListener, ILayerListener, City, GameState
from .DefaultGameMode import DefaultGameMode
from ..Mouse import Mouse
from ..component.frame import CityFrame, ResourcesFrame
from ..component.world import ResourcesComponent
from ..layer import SelectionValue, ShadowValue
from ..theme.Theme import Theme


class CityGameMode(DefaultGameMode, IGameStateListener, ILayerListener):

    def __init__(self, theme: Theme, logic: Logic, cityCell: Tuple[int, int], previousView: Tuple[int, int] = (0, 0)):
        super().__init__(theme, logic)

        self.__cityCell = cityCell
        self.__city = cast(City, logic.world.buildings.getItem(cityCell))
        self.__previousView = previousView

        # Selection
        self.__targetAction = ""
        self.__targetCell: Optional[Tuple[int, int]] = None
        self.__selectedCell: Optional[Tuple[int, int]] = None
        self._selectionLayer.setValue(cityCell, SelectionValue.ASSIGNED2)

        # Shadow
        rules = logic.rules
        self._shadowLayer.fill(ShadowValue.SHADOW_MEDIUM)
        pos, cityArea = rules.getCityArea(cityCell, ShadowValue.SHADOW_LIGHT, ShadowValue.SHADOW_MEDIUM)
        self._shadowLayer.blit(pos, cityArea)
        pos, houseArea = rules.getCityHouseArea(cityCell, ShadowValue.NONE, ShadowValue.SHADOW_LIGHT)
        self._shadowLayer.blit(pos, houseArea)

        # Components
        self._worldComponent.buildingsComponent.showAssigned = True
        self.__resourcesComponent = ResourcesComponent(theme, logic.rules, cityCell)
        self.addComponent(self.__resourcesComponent, cache=True)

        self.__resourcesFrame = ResourcesFrame(theme, logic)
        self.__resourcesFrame.moveRelativeTo("topLeft", self, "topLeft")
        self.addComponent(self.__resourcesFrame, cache=True)

        self.__cityFrame = CityFrame(theme, logic, self.__city)
        self.__cityFrame.moveRelativeTo("topLeft", self.__resourcesFrame, "bottomLeft")
        self.addComponent(self.__cityFrame, cache=True)

        # Listener
        self._state.registerListener(self)
        self._world.units.registerListener(self)

    def dispose(self):
        super().dispose()
        self._state.removeListener(self)
        self._world.units.removeListener(self)

    @property
    def previousView(self) -> Tuple[int, int]:
        return self.__previousView

    # GameState listener

    def citizenAssigned(self, state: GameState, cityCell: Tuple[int, int], cell: Tuple[int, int]):
        resources = self._logic.rules.computeCellResources(cityCell, cell)
        if cell == self.__targetCell:
            self.__updateTarget(cell)

    def citizenUnassigned(self, state: GameState, cityCell: Tuple[int, int], cell: Tuple[int, int]):
        if cell == self.__targetCell:
            self.__updateTarget(cell)

    # Component listener

    def worldCellEntered(self, cell: Tuple[int, int], mouse: Mouse, dragging: bool):
        self.__updateTarget(cell)

    def worldCellClicked(self, cell: Tuple[int, int], mouse: Mouse):
        if mouse.button1:
            command: Optional[Command] = None
            if self.__targetAction == "assignCitizen":
                command = AssignCitizen(self.__cityCell, cell)
                self._logic.addCommand(command)
            elif self.__targetAction == "unassignCitizen":
                command = UnassignCitizen(self.__cityCell, cell)
                self._logic.addCommand(command)
        elif mouse.button3:
            from .WorldGameMode import WorldGameMode
            gameMode = WorldGameMode(self.theme, self._logic)
            gameMode.viewChanged(self.__previousView)
            self.notifySetGameModeRequested(gameMode)

    def __clearTarget(self):
        if self.__targetCell is not None:
            self._selectionLayer.setValue(self.__targetCell, SelectionValue.NONE)
            self._selectionLayer.notifyCellChanged(self.__targetCell)
            self.__targetCell = None
            self.__targetAction = ""
        self._selectionLayer.clearNumbers()

    def __updateTarget(self, cell: Tuple[int, int]):
        self.__clearTarget()

        assignCommand = AssignCitizen(self.__cityCell, cell)
        if assignCommand.check(self._logic):
            self.__targetCell = cell
            self.__targetAction = "assignCitizen"
            self._selectionLayer.setValue(cell, SelectionValue.ASSIGN)
            self._selectionLayer.notifyCellChanged(cell)
            return

        unassignCommand = UnassignCitizen(self.__cityCell, cell)
        if unassignCommand.check(self._logic):
            self.__targetCell = cell
            self.__targetAction = "unassignCitizen"
            self._selectionLayer.setValue(cell, SelectionValue.UNASSIGN)
            self._selectionLayer.notifyCellChanged(cell)
            return
