from typing import Tuple

import pygame

from core.logic import Logic
from core.state import World
from tools.vector import vectorMulI, vectorSubI, vectorClampI
from .GameMode import GameMode
from ..component.IComponentListener import IComponentListener
from ..component.frame.MinimapFrame import MinimapFrame
from ..component.world.WorldComponent import WorldComponent
from ..theme.Theme import Theme


class DefaultGameMode(GameMode, IComponentListener):

    def __init__(self, theme: Theme, logic: Logic):
        super().__init__(theme)
        self._logic = logic
        state = logic.state
        self._state = state
        world = logic.world
        self._world = world
        self.__font = theme.getFont("default")

        # Mouse cursor
        tile = theme.getMouseCursor(state.playerId)
        cursor = pygame.cursors.Cursor((0, 0), tile)
        pygame.mouse.set_cursor(cursor)

        # Components
        self._worldComponent = WorldComponent(theme, logic)
        self._selectionLayer = self._worldComponent.selectionLayer
        self._shadowLayer = self._worldComponent.shadowLayer

        self._minimapFrame = MinimapFrame(theme, world)
        self._minimapFrame.moveRelativeTo("bottomRight", self, "bottomRight")

        # Components order
        self.addComponent(self._worldComponent)
        self.addComponent(self._minimapFrame)

        # Listener
        self._worldComponent.registerListener(self)  # worldCellClicked, ...
        self._minimapFrame.registerListener(self)  # viewChanged

    @property
    def world(self) -> World:
        return self._world

    @property
    def view(self) -> Tuple[int, int]:
        return self._worldComponent.view

    # Game loop

    def dispose(self):
        super().dispose()
        self._worldComponent.removeListener(self)
        self._minimapFrame.removeListener(self)

    def update(self):
        self._logic.executeCommands()
        super().update()

    # UI Event handler

    def processInput(self) -> bool:
        # Update view using keyboard state
        keys = pygame.key.get_pressed()
        newViewX, newViewY = self._worldComponent.view
        tileset = self.theme.getTileset("ground")
        tileWidth, tileHeight = tileset.tileSize
        speedX = tileWidth * 2
        speedY = tileHeight * 2
        if keys[pygame.K_RIGHT]:
            newViewX += speedX
        if keys[pygame.K_LEFT]:
            newViewX -= speedX
        if keys[pygame.K_UP]:
            newViewY -= speedY
        if keys[pygame.K_DOWN]:
            newViewY += speedY
        newView = (newViewX, newViewY)

        # Clamp new view
        worldSize = vectorMulI(self._world.size, tileset.tileSize)
        maxView = vectorSubI(worldSize, self.theme.viewSize)
        newView = vectorClampI(newView, 0, maxView)

        # Update only if changes
        if newView != self._worldComponent.view:
            self.viewChanged(newView)
            return True
        return False
