from typing import Tuple, Optional, cast

import pygame

from core.constants import ItemProperty, Direction, CellValue
from core.constants.Direction import transformCoords
from core.logic import Logic, DistanceMap
from core.logic.command import MoveUnit, AttackUnit
from core.state import Unit, IGameStateListener, GameState, ILayerListener, Layer, City
from .DefaultGameMode import DefaultGameMode
from ..Mouse import Mouse
from ..component.frame import TurnFrame, ResourcesFrame, UnitFrame
from ..layer import SelectionValue
from ..theme.Theme import Theme
from ..tools import centerCellView


class WorldGameMode(DefaultGameMode, IGameStateListener, ILayerListener):

    def __init__(self, theme: Theme, logic: Logic):
        super().__init__(theme, logic)

        # Selection
        self.__targetAction = ""
        self.__targetCell: Optional[Tuple[int, int]] = None
        self.__targetSize: Optional[Tuple[int, int]] = None
        self.__selectedCell: Optional[Tuple[int, int]] = None
        self.__distanceMap: Optional[DistanceMap] = None

        # Components
        self.__resourcesFrame = ResourcesFrame(theme, logic)
        self.__resourcesFrame.moveRelativeTo("topLeft", self, "topLeft")

        self.__unitFrame: Optional[UnitFrame] = None

        self.__turnFrame = TurnFrame(theme, logic)
        self.__turnFrame.moveRelativeTo("bottomRight", self._minimapFrame, "topRight")

        self.addComponent(self.__resourcesFrame, cache=True)
        self.addComponent(self.__turnFrame, cache=True)

        # Listener
        self._state.registerListener(self)
        self._world.units.registerListener(self)

    def dispose(self):
        super().dispose()
        self._state.removeListener(self)
        self._world.units.removeListener(self)

    # UI Event handler

    def keyDown(self, key: int) -> bool:
        if key in [pygame.K_e, pygame.K_F2]:
            from .EditGameMode import EditGameMode
            gameMode = EditGameMode(self.theme, self._logic)
            gameMode.viewChanged(self.view)
            self.notifySetGameModeRequested(gameMode)
            return True
        elif key in [pygame.K_F8]:  # Debug / Illustrate
            self._worldComponent.selectionComponent.rollShowMode()
            return True
        return False

    def mouseLeave(self) -> bool:
        self.__clearTarget()
        return True

    # Component listener

    def worldCellEntered(self, cell: Tuple[int, int], mouse: Mouse, dragging: bool):
        self.__updateTarget(cell)
        #print("Entered", cell)

    def worldCellClicked(self, cell: Tuple[int, int], mouse: Mouse):
        if mouse.button1:
            if self.__selectedCell == cell:
                return
            if self.__targetCell is None:
                return
            if self.__targetSize is not None:
                found = False
                rangeX, rangeY = self.__targetSize
                for sizeY in range(rangeY):
                    for sizeX in range(rangeX):
                        if (self.__targetCell[0] + sizeX) == cell[0] \
                                and (self.__targetCell[1] + sizeY) == cell[1]:
                            found = True
                if not found:
                    return
            if self.__targetAction == "selectUnit":
                self.__clearTarget()
                self.__selectUnit(cell)
            elif self.__targetAction == "moveUnit":
                self.__clearTarget()
                self.__moveUnit(cell)
            elif self.__targetAction == "attackUnit":
                self.__attackUnit(cell)
            elif self.__targetAction == "selectCity":
                self.__selectCity(cell)
            else:
                self.__clearSelection()
        elif mouse.button3:
            self.__clearSelection()
            self.__updateTarget(cell)

    def __clearTarget(self):
        if self.__targetCell is not None:
            self._selectionLayer.setValue(self.__targetCell, SelectionValue.NONE)
            self._selectionLayer.notifyCellChanged(self.__targetCell)
            self.__targetCell = None
            self.__targetSize = None
            self.__targetAction = ""
        self._selectionLayer.clearNumbers()

    def __updateTarget(self, cell: Tuple[int, int]):
        self.__clearTarget()

        if self.__selectedCell is not None:
            if cell == self.__selectedCell:
                return

            attackCommand = AttackUnit(self.__selectedCell, cell)
            if attackCommand.check(self._logic):
                self.__targetCell = cell
                self.__targetSize = (1, 1)
                self.__targetAction = "attackUnit"
                self._selectionLayer.setValue(cell, SelectionValue.ATTACK)
                self._selectionLayer.notifyCellChanged(cell)
                return

            moveCommand = MoveUnit(self.__selectedCell, cell, self.__distanceMap)
            if moveCommand.check(self._logic):
                self.__targetCell = cell
                self.__targetSize = (1, 1)
                self.__targetAction = "moveUnit"
                self._selectionLayer.showPath(moveCommand.path, moveCommand.costs)
                self._selectionLayer.notifyContentChanged()
                return

        unit = self.world.units.getItem(cell)
        if unit is not None and isinstance(unit, Unit):
            self.__targetCell = cell
            self.__targetSize = (1, 1)
            self.__targetAction = "selectUnit"
            self._selectionLayer.setValue(cell, SelectionValue.SELECT)
            self._selectionLayer.notifyCellChanged(cell)
            return

        buildings = self.world.buildings
        value = buildings.getValue(cell)
        if value == CellValue.BUILDINGS_CITY:
            mainCell = self._logic.rules.getCityTopLeft(cell)
            self.__targetCell = mainCell
            self.__targetSize = (2, 2)
            self.__targetAction = "selectCity"
            self._selectionLayer.setValue(mainCell, SelectionValue.SELECT2)
            self._selectionLayer.notifyCellChanged(mainCell)
            return

    def __clearSelection(self):
        if self.__selectedCell is not None:
            self._selectionLayer.clear()
            self._selectionLayer.notifyContentChanged()
            self.__selectedCell = None
            self._shadowLayer.clear()
            self._shadowLayer.notifyContentChanged()

        if self.__unitFrame is not None:
            self.removeComponent(self.__unitFrame)

        # Debug / Illustrate
        self._worldComponent.selectionComponent.setDistanceMap(None)

    def __selectUnit(self, cell: Tuple[int, int]):
        if self.__selectedCell is not None:
            self._selectionLayer.setValue(self.__selectedCell, SelectionValue.NONE)
            self._selectionLayer.clearNumbers()
        item = self.world.units.getItem(cell)
        if item is None:
            return
        unit = cast(Unit, item)
        self.__selectedCell = cell
        self._selectionLayer.setValue(cell, SelectionValue.SELECTED)
        self._selectionLayer.notifyCellChanged(cell)

        # Unit frame
        if self.__unitFrame is not None:
            self.removeComponent(self.__unitFrame)
        self.__unitFrame = UnitFrame(self.theme, self._logic, unit)
        self.__unitFrame.moveRelativeTo("bottomRight", self._minimapFrame, "bottomLeft")
        self.addComponent(self.__unitFrame)

        # Distance map
        x, y = cell
        radius = 31
        area = (
            x - radius, y - radius,
            x + radius + 1, y + radius + 1
        )
        self.__distanceMap = DistanceMap(self.world, area)
        self.__distanceMap.compute(cell)

        # Move area
        movePoints = unit.getIntProperty(ItemProperty.MOVE_POINTS)
        self._shadowLayer.showPaths(self.__distanceMap, movePoints)
        self._shadowLayer.notifyContentChanged()

        # Debug / Illustrate
        self._worldComponent.selectionComponent.setDistanceMap(self.__distanceMap)

    def __moveUnit(self, targetCell: Tuple[int, int]):
        if self.__selectedCell is None:
            return
        command = MoveUnit(self.__selectedCell, targetCell)
        self._logic.addCommand(command)

    def __attackUnit(self, cell: Tuple[int, int]):
        if self.__selectedCell is None:
            return
        command = AttackUnit(self.__selectedCell, cell)
        self._logic.addCommand(command)

    def __selectCity(self, cell: Tuple[int, int]):
        from .CityGameMode import CityGameMode
        cell = self._logic.rules.getCityTopLeft(cell)
        gameMode = CityGameMode(self.theme, self._logic, cell, self.view)
        view = centerCellView(self.theme, self.world, (cell[0] + 1, cell[1] + 1))
        gameMode.viewChanged(view)
        self.notifySetGameModeRequested(gameMode)

    # GameState listener

    def turnChanged(self, state: GameState):
        self.__clearSelection()
        tile = self.theme.getMouseCursor(state.playerId)
        cursor = pygame.cursors.Cursor((0, 0), tile)
        pygame.mouse.set_cursor(cursor)

    def unitMoved(self, state: GameState,
                  fromCell: Tuple[int, int], toCell: Tuple[int, int]):
        self.__selectUnit(toCell)

    def unitDamaged(self, state: GameState, cell: Tuple[int, int], damage: int):
        self._worldComponent.animationComponent.addDamage(cell, damage)

    def unitChanged(self, state: GameState, cell: Tuple[int, int], unit: Unit):
        if cell == self.__targetCell:
            self.__updateTarget(cell)

    def unitDied(self, state: GameState, cell: Tuple[int, int]):
        if cell == self.__targetCell:
            self.__updateTarget(cell)
        self._worldComponent.animationComponent.addGrave(cell)


